"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = __importDefault(require("assert"));
const nanoid_1 = __importDefault(require("nanoid"));
const ut = __importStar(require("./util"));
class figure_t {
    constructor(array) {
        this.array = array || new Array();
    }
    vertex(src, tar) {
        if (!this.has_src(src)) {
            this.array.push({ src, tar, cfg: new vertex_t() });
        }
        else {
            throw new Error(`src already defined`);
        }
    }
    edge(src, tar) {
        let cfg = new cell_t({
            dom: src.endpoints,
            cod: tar.endpoints,
        });
        cfg.fig.vertex(src.endpoints.vertex("start"), tar.endpoints.vertex("start"));
        cfg.fig.vertex(src.endpoints.vertex("end"), tar.endpoints.vertex("end"));
        if (!this.has_src(src)) {
            this.array.push({ src, tar, cfg });
        }
        else {
            throw new Error(`src already defined`);
        }
    }
    edge_rev(src, tar) {
        let cfg = new cell_t({
            dom: src.endpoints,
            cod: tar.endpoints,
        });
        cfg.fig.vertex(src.endpoints.vertex("start"), tar.endpoints.vertex("end"));
        cfg.fig.vertex(src.endpoints.vertex("end"), tar.endpoints.vertex("start"));
        if (!this.has_src(src)) {
            this.array.push({ src, tar, cfg });
        }
        else {
            throw new Error(`src already defined`);
        }
    }
    has_src(cell) {
        for (let { src } of this.array) {
            if (cell.eq(src)) {
                return true;
            }
        }
        return false;
    }
    has_tar(cell) {
        for (let { tar } of this.array) {
            if (cell.eq(tar)) {
                return true;
            }
        }
        return false;
    }
    repr() {
        let repr = [];
        for (let { src, tar, cfg } of this.array) {
            repr.push({
                src: src.uuid,
                tar: tar.uuid,
                cfg: cfg.repr(),
            });
        }
        return repr;
    }
    static from_polygon_and_circuit(polygon, cod, circuit) {
        let size = polygon.size;
        let fig = new figure_t();
        for (let i = 0; i < size; i++) {
            let src = polygon.edge_array[i];
            let edge = circuit[i];
            if (edge instanceof edge_t) {
                let tar = edge;
                fig.vertex(src.end, tar.end);
                fig.edge(src, tar);
            }
            else {
                let tar = edge.rev;
                fig.vertex(src.end, tar.start);
                fig.edge_rev(src, tar);
            }
        }
        return fig;
    }
    eq(that) {
        return ut.array_eq(this.array, that.array, (x, y) => x.src.eq(y.src) &&
            x.tar.eq(y.tar) &&
            x.cfg.eq(y.cfg));
    }
}
exports.figure_t = figure_t;
class morphism_t {
    constructor(the) {
        this.dom = the.dom;
        this.cod = the.cod;
        this.fig = the.fig || new figure_t();
        this.uuid = the.uuid || nanoid_1.default(10);
    }
    eq(that) {
        return (this.uuid === that.uuid) &&
            this.dom.eq(that.dom) &&
            this.cod.eq(that.cod) &&
            this.fig.eq(that.fig);
    }
    repr() {
        let repr = {};
        repr["uuid"] = this.uuid;
        repr["dom"] = this.dom.repr();
        repr["cod"] = this.cod.repr();
        repr["fig"] = this.fig.repr();
        return repr;
    }
}
exports.morphism_t = morphism_t;
class cell_t extends morphism_t {
    constructor(the) {
        super(the);
        this.dom = new spherical_t(the.dom, the.spherical_evidence);
    }
    get dim() {
        return this.dom.dim + 1;
    }
    ser(com) {
        for (let [i, cell] of com.cell_array(this.dim).entries()) {
            if (this.eq(cell)) {
                return i;
            }
        }
        throw new Error(`fail to find cell in com`);
    }
}
exports.cell_t = cell_t;
class cell_complex_t {
    constructor(the) {
        this.dim = the.dim;
        this.cell_map = the.cell_map || new Map();
        for (let d of ut.range(0, this.dim + 1)) {
            if (!this.cell_map.has(d)) {
                this.cell_map.set(d, new Array());
            }
        }
        this.name_map = the.name_map || new Map();
    }
    eq(that) {
        return (this.dim === that.dim) &&
            ut.map_eq(this.cell_map, that.cell_map, (x, y) => ut.array_eq(x, y, (v, w) => v.eq(w))) &&
            ut.map_eq(this.name_map, that.name_map, (x, y) => x.eq(y));
    }
    skeleton(dim) {
        let com = new cell_complex_t({ dim });
        for (let [d, cell_array] of this.cell_map) {
            if (d <= dim) {
                com.cell_map.set(d, cell_array);
            }
        }
        for (let [name, cell] of this.name_map) {
            if (cell.dim <= dim) {
                com.name_map.set(name, cell);
            }
        }
        return com;
    }
    cell_array(dim) {
        let cell_array = this.cell_map.get(dim);
        if (cell_array !== undefined) {
            return cell_array;
        }
        else {
            return new Array();
        }
    }
    dim_size(dim) {
        return this.cell_array(dim).length;
    }
    *cells() {
        for (let d of ut.range(0, this.dim + 1)) {
            for (let cell of this.cell_array(d)) {
                yield cell;
            }
        }
    }
    get vertex_array() {
        return this.cell_array(0).map(cell => cell);
    }
    get edge_array() {
        return this.cell_array(1).map(cell => cell);
    }
    has_name(name) {
        return this.name_map.has(name);
    }
    has_cell(x) {
        for (let y of this.cells()) {
            if (x.eq(y)) {
                return true;
            }
        }
        return false;
    }
    attach(name, cell) {
        this.cell_array(cell.dim).push(cell);
        this.name_map.set(name, cell);
    }
    attach_vertex(name) {
        this.attach(name, new vertex_t());
    }
    attach_vertexes(names) {
        for (let name of names) {
            this.attach_vertex(name);
        }
    }
    attach_edge(name, [start, end]) {
        this.attach(name, new edge_t(this, this.vertex(start), this.vertex(end)));
    }
    attach_face(name, circuit_exp) {
        this.attach(name, new face_t(this, circuit_exp.map(exp => edge_exp_eval(this, exp))));
    }
    cell(name) {
        let cell = this.name_map.get(name);
        if (cell !== undefined) {
            return cell;
        }
        else {
            throw new Error(`can not find cell: ${name}`);
        }
    }
    vertex(name) {
        let cell = this.cell(name);
        if (cell instanceof vertex_t) {
            return cell;
        }
        else {
            throw new Error(`cell_t not vertex_t: ${name}`);
        }
    }
    edge(name) {
        let cell = this.cell(name);
        if (cell instanceof edge_t) {
            return cell;
        }
        else {
            throw new Error(`cell_t not edge_t: ${name}`);
        }
    }
    face(name) {
        let cell = this.cell(name);
        if (cell instanceof face_t) {
            return cell;
        }
        else {
            throw new Error(`cell_t not edge_t: ${name}`);
        }
    }
    empty_p() {
        return this.dim === -1;
    }
    repr() {
        if (this.empty_p()) {
            return null;
        }
        let repr = {};
        repr["dim"] = this.dim;
        for (let [d, cell_array] of this.cell_map) {
            repr[d] = cell_array.map(cell => cell.repr());
        }
        repr["name_map"] = {};
        for (let [name, cell] of this.name_map) {
            repr["name_map"][name] = cell.uuid;
        }
        return repr;
    }
}
exports.cell_complex_t = cell_complex_t;
class isomorphism_t extends morphism_t {
    constructor(the) {
        super(the);
        if (!epimorphism_p(this)) {
            ut.log(this.repr());
            throw new Error("not epimorphism");
        }
        if (!monomorphism_p(this)) {
            ut.log(this.repr());
            throw new Error("not monomorphism");
        }
    }
    static endpoints(com) {
        if (com.dim !== 0) {
            return undefined;
        }
        let size = com.vertex_array.length;
        if (size !== 2) {
            return undefined;
        }
        let [start, end] = com.vertex_array;
        let endpoints = new endpoints_t();
        let fig = new figure_t();
        fig.vertex(endpoints.vertex("start"), start);
        fig.vertex(endpoints.vertex("end"), end);
        return new isomorphism_t({
            dom: endpoints,
            cod: com,
            fig,
        });
    }
    static polygon(com) {
        if (com.dim !== 1) {
            return undefined;
        }
        let size = com.edge_array.length;
        if (size !== com.vertex_array.length) {
            return undefined;
        }
        if (size === 0) {
            return undefined;
        }
        let circuit = new Array();
        let vertex = com.vertex_array[0];
        let edge_uuid_set = new Set();
        for (let i = 0; i < size; i++) {
            for (let edge of com.edge_array) {
                if (!edge_uuid_set.has(edge.uuid)) {
                    if (edge.start.eq(vertex)) {
                        edge_uuid_set.add(edge.uuid);
                        circuit.push(edge);
                        vertex = edge.end;
                        break;
                    }
                    else if (edge.end.eq(vertex)) {
                        edge_uuid_set.add(edge.uuid);
                        circuit.push(edge);
                        vertex = edge.start;
                        break;
                    }
                }
            }
        }
        if (circuit.length !== size) {
            return undefined;
        }
        let polygon = new polygon_t(size);
        let fig = figure_t
            .from_polygon_and_circuit(polygon, com, circuit);
        return new isomorphism_t({
            dom: polygon,
            cod: com,
            fig,
        });
    }
}
exports.isomorphism_t = isomorphism_t;
class epimorphism_t extends morphism_t {
    constructor(the) {
        super(the);
        if (!epimorphism_p(this)) {
            throw new Error("not epimorphism");
        }
    }
}
exports.epimorphism_t = epimorphism_t;
class monomorphism_t extends morphism_t {
    constructor(the) {
        super(the);
        if (!monomorphism_p(this)) {
            throw new Error("not monomorphism");
        }
    }
}
exports.monomorphism_t = monomorphism_t;
function isomorphism_p(the) {
    return epimorphism_p(the) && monomorphism_p(the);
}
exports.isomorphism_p = isomorphism_p;
function epimorphism_p(the) {
    for (let cell of the.cod.cells()) {
        if (!the.fig.has_tar(cell)) {
            return false;
        }
    }
    return true;
}
exports.epimorphism_p = epimorphism_p;
function monomorphism_p(the) {
    let used_uuid_set = new Set();
    for (let { src, tar } of the.fig.array) {
        if (used_uuid_set.has(tar.uuid)) {
            return false;
        }
        else {
            used_uuid_set.add(tar.uuid);
        }
    }
    return true;
}
exports.monomorphism_p = monomorphism_p;
function spherical_check(evid, com) {
    return evid.iso.cod.eq(com);
}
exports.spherical_check = spherical_check;
class spherical_t extends cell_complex_t {
    constructor(com, evid) {
        super(com);
        this.spherical_evidence = evid ||
            spherical_evidence_t.generate(this) ||
            ut.panic("spherical_evidence_t.generate fail");
        assert_1.default(spherical_check(this.spherical_evidence, com));
    }
}
exports.spherical_t = spherical_t;
class spherical_evidence_t {
    constructor(iso) {
        this.iso = iso;
    }
    get dim() {
        return this.iso.dom.dim;
    }
    static from_iso(iso) {
        if (iso !== undefined) {
            return new spherical_evidence_t(iso);
        }
        else {
            return undefined;
        }
    }
    static generate(com) {
        if (com.empty_p()) {
            return new spherical_evidence_t(new isomorphism_t(new empty_morphism_t()));
        }
        else if (com.dim === 0) {
            return spherical_evidence_t.from_iso(isomorphism_t.endpoints(com));
        }
        else if (com.dim === 1) {
            return spherical_evidence_t.from_iso(isomorphism_t.polygon(com));
        }
        else {
            console.log("[warning] can not check dim 2 yet");
            return undefined;
        }
    }
}
exports.spherical_evidence_t = spherical_evidence_t;
// export
// class manifold_t extends cell_complex_t {
//   manifold_evidence: manifold_evidence_t
//   constructor (
//     com: cell_complex_t,
//     evid?: manifold_evidence_t,
//   ) {
//     super (com)
//     if (evid !== undefined) {
//       if (evid.iso.cod.eq (this)) {
//         this.manifold_evidence = evid
//       } else {
//         throw new Error ("manifold_check fail")
//       }
//     } else {
//       evid = manifold_evidence_t.generate (this)
//       if (evid !== undefined) {
//         this.manifold_evidence = evid
//       } else {
//         throw new Error ("manifold_evidence_t.generate fail")
//       }
//     }
//   }
// }
/**
 * A n-dim manifold, is a topological space,
 * each vertex of which has a n-ball as close neighbourhood.

 * For cell-complex, it is sufficient to check that,
 * each vertexes of it has a (n-1)-sphere as [[vertex_figure_t]].
 */
class manifold_evidence_t {
    // iso_map: Map <cell_t, isomorphism_t>
    constructor(dim) {
        this.dim = dim;
        // this.iso_map = new Map ()
    }
}
exports.manifold_evidence_t = manifold_evidence_t;
//// -1 dimension
class empty_complex_t extends cell_complex_t {
    constructor() {
        super({ dim: -1 });
    }
}
exports.empty_complex_t = empty_complex_t;
//// 0 dimension
class empty_morphism_t extends morphism_t {
    constructor() {
        super({
            dom: new empty_complex_t(),
            cod: new empty_complex_t(),
        });
    }
}
exports.empty_morphism_t = empty_morphism_t;
class vertex_t extends cell_t {
    constructor() {
        super({
            dom: new empty_complex_t(),
            cod: new empty_complex_t(),
        });
    }
}
exports.vertex_t = vertex_t;
class endpoints_t extends cell_complex_t {
    constructor() {
        super({ dim: 0 });
        this.attach_vertexes(["start", "end"]);
    }
}
exports.endpoints_t = endpoints_t;
//// 1 dimension
class edge_t extends cell_t {
    constructor(com, start, end) {
        let dom = new endpoints_t();
        let cod = com.skeleton(0);
        let fig = new figure_t();
        fig.vertex(dom.vertex("start"), start);
        fig.vertex(dom.vertex("end"), end);
        super({ dom, cod, fig });
        this.start = start;
        this.end = end;
        this.endpoints = dom;
    }
    get rev() {
        return new edge_rev_t(this);
    }
}
exports.edge_t = edge_t;
class polygon_t extends cell_complex_t {
    constructor(size) {
        super({ dim: 1 });
        this.size = size;
        for (let i = 0; i < size; i++) {
            this.attach_vertex(`0:${i}`);
        }
        for (let i = 0; i < size - 1; i++) {
            this.attach_edge(`1:${i}`, [`0:${i}`, `0:${i + 1}`]);
        }
        this.attach_edge(`1:${size - 1}`, [`0:${size - 1}`, `0:0`]);
    }
}
exports.polygon_t = polygon_t;
//// 2 dimension
class edge_rev_t {
    constructor(rev) {
        this.rev = rev;
    }
}
exports.edge_rev_t = edge_rev_t;
function edge_exp_eval(com, exp) {
    if (typeof exp === "string") {
        let name = exp;
        return com.edge(name);
    }
    else if (exp.length === 2 && exp[1] === "rev") {
        let name = exp[0];
        return new edge_rev_t(com.edge(name));
    }
    else {
        throw new Error(`unknown exp: ${exp}`);
    }
}
exports.edge_exp_eval = edge_exp_eval;
class face_t extends cell_t {
    constructor(com, circuit) {
        let dom = new polygon_t(circuit.length);
        let cod = com.skeleton(1);
        let fig = figure_t
            .from_polygon_and_circuit(dom, cod, circuit);
        super({ dom, cod, fig });
        this.circuit = circuit;
        this.polygon = dom;
    }
}
exports.face_t = face_t;
//# sourceMappingURL=cell-complex.js.map