"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = __importStar(require("lodash"));
const assert_1 = __importDefault(require("assert"));
const ut = __importStar(require("./util"));
const permutation_1 = require("./permutation");
const group_1 = require("./abstract/group");
const ring_1 = require("./abstract/ring");
function argfirst(lo, hi, p) {
    for (let i = lo; i < hi; i++) {
        if (p(i)) {
            return i;
        }
    }
    return null;
}
exports.argfirst = argfirst;
function argall(lo, hi, p) {
    for (let i = lo; i < hi; i++) {
        if (!p(i)) {
            return false;
        }
    }
    return true;
}
exports.argall = argall;
function argsome(lo, hi, p) {
    for (let i = lo; i < hi; i++) {
        if (p(i)) {
            return true;
        }
    }
    return false;
}
exports.argsome = argsome;
function argcmp(lo, hi, c) {
    let arg = 0;
    for (let i = arg; i < hi; i++) {
        if (c(i, arg)) {
            arg = i;
        }
    }
    return arg;
}
exports.argcmp = argcmp;
function argcmp_guard(lo, hi, c, p) {
    let arg = argfirst(lo, hi, p);
    if (arg === null) {
        throw new Error("no such arg");
    }
    for (let i = arg + 1; i < hi; i++) {
        if (p(i)) {
            if (c(i, arg)) {
                arg = i;
            }
        }
    }
    return arg;
}
exports.argcmp_guard = argcmp_guard;
function ring(the) {
    return new ring_1.euclidean_ring_t({
        addition: new group_1.abelian_group_t({
            elements: the.elements,
            id: the.zero,
            add: the.add,
            neg: the.neg,
        }),
        multiplication: new group_1.monoid_t({
            elements: the.elements,
            id: the.one,
            mul: the.mul,
        }),
        degree_lt: the.degree_lt,
        divmod: the.divmod,
    });
}
exports.ring = ring;
class matrix_t {
    constructor(the) {
        this.ring = the.ring;
        this.buffer = the.buffer;
        this.shape = the.shape;
        this.strides = the.strides;
        this.offset = the.offset === undefined ? 0 : the.offset;
        this.size = matrix_t.shape_to_size(the.shape);
    }
    static shape_to_size([x, y]) {
        return x * y;
    }
    static shape_to_strides([x, y]) {
        return [y, 1];
    }
    static from_ring_buffer(ring, buffer, shape) {
        let strides = matrix_t.shape_to_strides(shape);
        return new matrix_t({ ring, buffer, shape, strides });
    }
    static from_ring_Array2d(ring, array) {
        let y_length = array.length;
        let x_length = array[0].length;
        for (let a of array) {
            if (a.length !== x_length) {
                throw new Error("inner array length mismatch");
            }
        }
        return matrix_t.from_ring_buffer(ring, Array.from(array.flat()), [y_length, x_length]);
    }
    static from_ring_row(ring, row) {
        let [s] = row.strides;
        return new matrix_t({
            ring,
            buffer: row.buffer,
            shape: [1, row.size],
            strides: [0, s],
            offset: row.offset,
        });
    }
    static from_ring_col(ring, col) {
        let [t] = col.strides;
        return new matrix_t({
            ring,
            buffer: col.buffer,
            shape: [col.size, 1],
            strides: [t, 0],
            offset: col.offset,
        });
    }
    get_linear_index(x, y) {
        return (this.offset +
            x * this.strides[0] +
            y * this.strides[1]);
    }
    get(x, y) {
        return this.buffer[this.get_linear_index(x, y)];
    }
    set(x, y, v) {
        this.buffer[this.get_linear_index(x, y)] = v;
        return this;
    }
    update_at(i, j, f) {
        this.set(i, j, f(this.get(i, j)));
        return this;
    }
    static *indexes_of_shape([m, n]) {
        for (let i = 0; i < m; i++) {
            for (let j = 0; j < n; j++) {
                yield [i, j];
            }
        }
    }
    *indexes() {
        for (let [i, j] of matrix_t.indexes_of_shape(this.shape)) {
            yield [i, j];
        }
    }
    *entries() {
        for (let [i, j] of this.indexes()) {
            let v = this.get(i, j);
            yield [i, j, v];
        }
    }
    *values() {
        for (let [i, j] of this.indexes()) {
            let v = this.get(i, j);
            yield v;
        }
    }
    update_op1(op) {
        let [m, n] = this.shape;
        for (let i = 0; i < m; i++) {
            for (let j = 0; j < n; j++) {
                let v = op(this.get(i, j));
                this.set(i, j, v);
            }
        }
        return this;
    }
    same_shape_p(that) {
        return ((this.shape[0] === that.shape[0]) &&
            (this.shape[1] === that.shape[1]));
    }
    update_op2(that, op) {
        assert_1.default(this.same_shape_p(that));
        let [m, n] = this.shape;
        for (let i = 0; i < m; i++) {
            for (let j = 0; j < n; j++) {
                let v = op(this.get(i, j), that.get(i, j));
                this.set(i, j, v);
            }
        }
        return this;
    }
    slice(x, y) {
        let [m, n] = this.shape;
        let offset = this.offset;
        if (x !== null) {
            let [start, end] = x;
            m = end - start;
            offset += start * this.strides[0];
        }
        if (y !== null) {
            let [start, end] = y;
            n = end - start;
            offset += start * this.strides[1];
        }
        return new matrix_t({
            ring: this.ring,
            buffer: this.buffer,
            shape: [m, n],
            strides: this.strides,
            offset,
        });
    }
    copy() {
        let matrix = matrix_t.from_ring_buffer(this.ring, new Array(this.size), this.shape);
        for (let [i, j, x] of this.entries()) {
            matrix.set(i, j, x);
        }
        return matrix;
    }
    update_copy(that) {
        let [m, n] = this.shape;
        for (let i of ut.range(0, m)) {
            for (let j of ut.range(0, n)) {
                this.set(i, j, that.get(i, j));
            }
        }
        return this;
    }
    row(i) {
        let [m, n] = this.shape;
        let [s, t] = this.strides;
        let offset = this.offset + i * s;
        return new vector_t({
            ring: this.ring,
            buffer: this.buffer,
            shape: [n],
            strides: [t],
            offset,
        });
    }
    set_row(i, src) {
        let [m, n] = this.shape;
        for (let j of ut.range(0, n)) {
            this.set(i, j, src.get(j));
        }
        return this;
    }
    *row_entries() {
        let [m, _n] = this.shape;
        for (let i = 0; i < m; i++) {
            yield [i, this.row(i)];
        }
    }
    *rows() {
        let [m, _n] = this.shape;
        for (let i = 0; i < m; i++) {
            yield this.row(i);
        }
    }
    for_each_row_index(f) {
        for (let [i, row] of this.row_entries()) {
            f(row, i);
        }
        return this;
    }
    for_each_row(f) {
        for (let row of this.rows()) {
            f(row);
        }
        return this;
    }
    col(j) {
        let [m, n] = this.shape;
        let [s, t] = this.strides;
        let offset = this.offset + j * t;
        return new vector_t({
            ring: this.ring,
            buffer: this.buffer,
            shape: [m],
            strides: [s],
            offset,
        });
    }
    set_col(j, src) {
        let [m, n] = this.shape;
        for (let i of ut.range(0, m)) {
            this.set(i, j, src.get(i));
        }
        return this;
    }
    *col_entries() {
        let [_m, n] = this.shape;
        for (let i = 0; i < n; i++) {
            yield [i, this.col(i)];
        }
    }
    *cols() {
        let [_m, n] = this.shape;
        for (let i = 0; i < n; i++) {
            yield this.col(i);
        }
    }
    for_each_col_index(f) {
        for (let [i, col] of this.col_entries()) {
            f(col, i);
        }
        return this;
    }
    for_each_col(f) {
        for (let col of this.cols()) {
            f(col);
        }
        return this;
    }
    reduce_with(init, f) {
        let acc = init;
        for (let v of this.values()) {
            acc = f(acc, v);
        }
        return acc;
    }
    every(p) {
        for (let v of this.values()) {
            if (!p(v)) {
                return false;
            }
        }
        return true;
    }
    some(p) {
        for (let v of this.values()) {
            if (p(v)) {
                return true;
            }
        }
        return false;
    }
    *zip(that) {
        let this_iter = this.values();
        let that_iter = that.values();
        while (true) {
            let this_next = this_iter.next();
            let that_next = that_iter.next();
            if (this_next.done || that_next.done) {
                return;
            }
            else {
                yield [this_next.value, that_next.value];
            }
        }
    }
    eq(that) {
        if (this.size !== that.size) {
            return false;
        }
        if (!_.isEqual(this.shape, that.shape)) {
            return false;
        }
        for (let [x, y] of this.zip(that)) {
            if (!this.ring.eq(x, y)) {
                return false;
            }
        }
        return true;
    }
    toArray2d() {
        let array = [];
        let [x, _] = this.shape;
        for (let i = 0; i < x; i++) {
            array.push(this.row(i).toArray());
        }
        return array;
    }
    print() {
        console.log("matrix:");
        console.table(this.toArray2d());
    }
    update_add(that) {
        return this.update_op2(that, (x, y) => this.ring.add(x, y));
    }
    add(that) {
        return this.copy().update_add(that);
    }
    update_sub(that) {
        return this.update_op2(that, (x, y) => this.ring.sub(x, y));
    }
    sub(that) {
        return this.copy().update_sub(that);
    }
    map(f) {
        let matrix = this.copy();
        for (let [i, j, v] of this.entries()) {
            matrix.update_at(i, j, f);
        }
        return matrix;
    }
    update_scale(a) {
        return this.update_op1(x => this.ring.mul(x, a));
    }
    scale(a) {
        return this.map(n => this.ring.mul(n, a));
    }
    mul(that) {
        let [m, n] = this.shape;
        let [p, q] = that.shape;
        assert_1.default(n === p);
        let shape = [m, q];
        let size = matrix_t.shape_to_size(shape);
        let buffer = new Array(size);
        let matrix = matrix_t.from_ring_buffer(this.ring, buffer, shape);
        for (let i of ut.range(0, m)) {
            for (let j of ut.range(0, q)) {
                let v = this.row(i).dot(that.col(j));
                matrix.set(i, j, v);
            }
        }
        return matrix;
    }
    act(v) {
        return v.trans(this);
    }
    transpose() {
        let [m, n] = this.shape;
        let [s, t] = this.strides;
        return new matrix_t({
            ring: this.ring,
            buffer: this.buffer,
            shape: [n, m],
            strides: [t, s],
            offset: this.offset,
        });
    }
    symmetric_p() {
        return this.eq(this.transpose());
    }
    square_p() {
        let [x, y] = this.shape;
        return x === y;
    }
    zero_p() {
        return this.every(x => this.ring.zero_p(x));
    }
    update_swap_rows(i, j) {
        let [m, n] = this.shape;
        for (let k of ut.range(0, n)) {
            let x = this.get(i, k);
            let y = this.get(j, k);
            this.set(i, k, y);
            this.set(j, k, x);
        }
        return this;
    }
    update_swap_cols(i, j) {
        let [m, n] = this.shape;
        for (let k of ut.range(0, m)) {
            let x = this.get(k, i);
            let y = this.get(k, j);
            this.set(k, i, y);
            this.set(k, j, x);
        }
        return this;
    }
    append_cols(that) {
        let [m, n] = this.shape;
        let [p, q] = that.shape;
        assert_1.default(m === p);
        let shape = [m, n + q];
        let size = matrix_t.shape_to_size(shape);
        let buffer = new Array(size);
        let matrix = matrix_t.from_ring_buffer(this.ring, buffer, shape);
        for (let i of ut.range(0, m)) {
            let row = this.row(i).append(that.row(i));
            matrix.set_row(i, row);
        }
        return matrix;
    }
    append_rows(that) {
        let [m, n] = this.shape;
        let [p, q] = that.shape;
        assert_1.default(n === q);
        let shape = [m + p, n];
        let size = matrix_t.shape_to_size(shape);
        let buffer = new Array(size);
        let matrix = matrix_t.from_ring_buffer(this.ring, buffer, shape);
        for (let i of ut.range(0, n)) {
            let col = this.col(i).append(that.col(i));
            matrix.set_col(i, col);
        }
        return matrix;
    }
    /**
     * The main diag.
     */
    diag() {
        let [m, n] = this.shape;
        n = Math.min(m, n);
        let vector = vector_t.ring_zeros(this.ring, n);
        for (let i of ut.range(0, n)) {
            vector.set(i, this.get(i, i));
        }
        return vector;
    }
    static ring_constant(ring, n, x, y) {
        let shape = [x, y];
        let size = matrix_t.shape_to_size(shape);
        let buffer = new Array(size);
        buffer.fill(n);
        return matrix_t.from_ring_buffer(ring, buffer, shape);
    }
    static ring_zeros(ring, x, y) {
        return matrix_t.ring_constant(ring, ring.zero, x, y);
    }
    static ring_ones(ring, x, y) {
        return matrix_t.ring_constant(ring, ring.one, x, y);
    }
    static ring_id(ring, n) {
        let matrix = matrix_t.ring_zeros(ring, n, n);
        for (let i of ut.range(0, n)) {
            matrix.set(i, i, ring.one);
        }
        return matrix;
    }
    static ring_row_trans(ring, permutation) {
        let n = permutation.size;
        let matrix = matrix_t.ring_zeros(ring, n, n);
        for (let [i, v] of permutation.pairs()) {
            matrix.set(i, v, ring.one);
        }
        return matrix;
    }
    static ring_col_trans(ring, permutation) {
        let n = permutation.size;
        let matrix = matrix_t.ring_zeros(ring, n, n);
        for (let [i, v] of permutation.pairs()) {
            matrix.set(v, i, ring.one);
        }
        return matrix;
    }
    tuck_row(i, j) {
        let [m, n] = this.shape;
        let row_trans = matrix_t.ring_row_trans(this.ring, permutation_1.permutation_t.id(m).tuck(i, j));
        return row_trans.mul(this);
    }
    tuck_col(i, j) {
        let [m, n] = this.shape;
        let col_trans = matrix_t.ring_col_trans(this.ring, permutation_1.permutation_t.id(n).tuck(i, j));
        return this.mul(col_trans);
    }
    update_tuck_row(i, j) {
        this.update_copy(this.tuck_row(i, j));
        return this;
    }
    update_tuck_col(i, j) {
        this.update_copy(this.tuck_col(i, j));
        return this;
    }
    /**
     * I preserve the terms
     * `hermite_normal_form` and `smith_normal_form`
     * for matrix_t of integers.
     */
    /**
     * Generic `hermite_normal_form`
     */
    row_canonical_form() {
        let matrix = this.copy();
        let [m, n] = this.shape;
        let i = 0;
        let j = 0;
        while (i < m && j < n) {
            if (argall(i, m, (k) => this.ring.zero_p(matrix.get(k, j)))) {
                j += 1;
            }
            else {
                while (!((argall(i + 1, m, (k) => this.ring.zero_p(matrix.get(k, j))) &&
                    argall(0, i, (k) => this.ring.degree_lt(matrix.get(k, j), matrix.get(i, j)))))) {
                    let k = argcmp_guard(i, m, (k, l) => this.ring.degree_lt(matrix.get(k, j), matrix.get(l, j)), (k) => !this.ring.zero_p(matrix.get(k, j)));
                    if (k !== i) {
                        matrix.update_swap_rows(i, k);
                    }
                    for (let k of ut.ranges([[0, i], [i + 1, m]])) {
                        let q = this.ring.div(matrix.get(k, j), matrix.get(i, j));
                        if (!this.ring.zero_p(q)) {
                            matrix.row(k)
                                .update_sub(matrix.row(i).scale(q));
                        }
                    }
                }
                i += 1;
                j += 1;
            }
        }
        return matrix;
    }
    /**
     * `row_trans.mul (this) .eq (row_canonical)`
     */
    row_canonical_decomposition() {
        let [m, n] = this.shape;
        let augmented = this.append_cols(matrix_t.ring_id(this.ring, m));
        let echelon = augmented.row_canonical_form();
        return {
            row_trans: echelon.slice(null, [n, n + m]),
            row_canonical: echelon.slice(null, [0, n]),
        };
    }
    /**
     * Generic `smith_normal_form`
     */
    diag_canonical_update() {
        let matrix = this;
        let [m, n] = this.shape;
        let i = 0;
        let j = 0;
        while (i < m && j < n) {
            if ((argall(i, m, (k) => this.ring.zero_p(matrix.get(k, j))) &&
                argall(j, n, (k) => this.ring.zero_p(matrix.get(i, k))))) {
                i += 1;
                j += 1;
            }
            else {
                // It is amazing that this loop converges.
                // It is like kneading dough.
                while (!((argall(i + 1, m, (k) => this.ring.zero_p(matrix.get(k, j))) &&
                    argall(j + 1, n, (k) => this.ring.zero_p(matrix.get(i, k)))))) {
                    while (!argall(i + 1, m, (k) => this.ring.zero_p(matrix.get(k, j)))) {
                        let k = argcmp_guard(i, m, (k, l) => this.ring.degree_lt(matrix.get(k, j), matrix.get(l, j)), (k) => !this.ring.zero_p(matrix.get(k, j)));
                        if (k !== i) {
                            matrix.update_swap_rows(i, k);
                        }
                        for (let k of ut.range(i + 1, m)) {
                            let q = this.ring.div(matrix.get(k, j), matrix.get(i, j));
                            if (!this.ring.zero_p(q)) {
                                matrix.row(k)
                                    .update_sub(matrix.row(i).scale(q));
                            }
                        }
                    }
                    while (!argall(j + 1, n, (k) => this.ring.zero_p(matrix.get(i, k)))) {
                        let k = argcmp_guard(j, n, (k, l) => this.ring.degree_lt(matrix.get(i, k), matrix.get(i, l)), (k) => !this.ring.zero_p(matrix.get(i, k)));
                        if (k !== j) {
                            matrix.update_swap_cols(j, k);
                        }
                        for (let k of ut.range(j + 1, n)) {
                            let q = this.ring.div(matrix.get(i, k), matrix.get(i, j));
                            if (!this.ring.zero_p(q)) {
                                matrix.col(k)
                                    .update_sub(matrix.col(j).scale(q));
                            }
                        }
                    }
                }
                i += 1;
                j += 1;
            }
        }
        return matrix;
    }
    invariant_factor_update() {
        let matrix = this;
        let [m, n] = this.shape;
        m = Math.min(m, n);
        let i = 0;
        while (i < m) {
            if (this.ring.zero_p(matrix.get(i, i))) {
                for (let k of ut.range(i, m)) {
                    matrix.set(k, k, matrix.get(k + 1, k + 1));
                }
                matrix.set(m - 1, m - 1, this.ring.zero);
                m -= 1;
            }
            else {
                let x = matrix.get(i, i);
                for (let k of ut.range(i + 1, m)) {
                    let y = matrix.get(k, k);
                    let [q, r] = this.ring.divmod(y, x);
                    if (!this.ring.zero_p(r)) {
                        matrix.set(k, i, matrix.get(k, k));
                        matrix.diag_canonical_update();
                    }
                }
                i += 1;
            }
        }
        return matrix;
    }
    diag_canonical_form() {
        return this.copy()
            .diag_canonical_update()
            .invariant_factor_update();
    }
    diagonal_p() {
        let [m, n] = this.shape;
        for (let i of ut.range(0, m)) {
            for (let j of ut.range(0, n)) {
                if (i !== j) {
                    if (!this.ring.zero_p(this.get(i, j))) {
                        return false;
                    }
                }
            }
        }
        return true;
    }
    diag_canonical_form_p() {
        if (!this.diagonal_p()) {
            return false;
        }
        if (!this.diag().invariant_factors_p()) {
            return false;
        }
        return true;
    }
    diag_canonical_update_ext(the) {
        let matrix = this;
        let [m, n] = this.shape;
        let row_trans = the.row_trans;
        let col_trans = the.col_trans;
        let i = 0;
        let j = 0;
        while (i < m && j < n) {
            if ((argall(i, m, (k) => this.ring.zero_p(matrix.get(k, j))) &&
                argall(j, n, (k) => this.ring.zero_p(matrix.get(i, k))))) {
                i += 1;
                j += 1;
            }
            else {
                while (!((argall(i + 1, m, (k) => this.ring.zero_p(matrix.get(k, j))) &&
                    argall(j + 1, n, (k) => this.ring.zero_p(matrix.get(i, k)))))) {
                    while (!argall(i + 1, m, (k) => this.ring.zero_p(matrix.get(k, j)))) {
                        let k = argcmp_guard(i, m, (k, l) => this.ring.degree_lt(matrix.get(k, j), matrix.get(l, j)), (k) => !this.ring.zero_p(matrix.get(k, j)));
                        if (k !== i) {
                            matrix.update_swap_rows(i, k);
                            row_trans.update_swap_rows(i, k);
                        }
                        for (let k of ut.range(i + 1, m)) {
                            let q = this.ring.div(matrix.get(k, j), matrix.get(i, j));
                            if (!this.ring.zero_p(q)) {
                                matrix.row(k)
                                    .update_sub(matrix.row(i).scale(q));
                                row_trans.row(k)
                                    .update_sub(row_trans.row(i).scale(q));
                            }
                        }
                    }
                    while (!argall(j + 1, n, (k) => this.ring.zero_p(matrix.get(i, k)))) {
                        let k = argcmp_guard(j, n, (k, l) => this.ring.degree_lt(matrix.get(i, k), matrix.get(i, l)), (k) => !this.ring.zero_p(matrix.get(i, k)));
                        if (k !== j) {
                            matrix.update_swap_cols(j, k);
                            col_trans.update_swap_cols(j, k);
                        }
                        for (let k of ut.range(j + 1, n)) {
                            let q = this.ring.div(matrix.get(i, k), matrix.get(i, j));
                            if (!this.ring.zero_p(q)) {
                                matrix.col(k)
                                    .update_sub(matrix.col(j).scale(q));
                                col_trans.col(k)
                                    .update_sub(col_trans.col(j).scale(q));
                            }
                        }
                    }
                }
                i += 1;
                j += 1;
            }
        }
    }
    invariant_factor_update_ext(the) {
        let [m, n] = this.shape;
        m = Math.min(m, n);
        let i = 0;
        while (i < m) {
            if (this.ring.zero_p(this.get(i, i))) {
                this.update_tuck_row(i, m - 1);
                this.update_tuck_col(i, m - 1);
                the.row_trans.update_tuck_row(i, m - 1);
                the.col_trans.update_tuck_col(i, m - 1);
                m -= 1;
            }
            else {
                let x = this.get(i, i);
                for (let k of ut.range(i + 1, m)) {
                    let y = this.get(k, k);
                    let [q, r] = this.ring.divmod(y, x);
                    if (!this.ring.zero_p(r)) {
                        this.col(i)
                            .update_add(this.col(k));
                        the.col_trans.col(i)
                            .update_add(the.col_trans.col(k));
                        this.diag_canonical_update_ext(the);
                    }
                }
                i += 1;
            }
        }
    }
    /**
     * `row_trans.mul (this) .mul (col_trans) .eq (diag_canonical)`
     */
    diag_canonical_decomposition() {
        let [m, n] = this.shape;
        let diag_canonical = this.copy();
        let the = {
            row_trans: matrix_t.ring_id(this.ring, m),
            col_trans: matrix_t.ring_id(this.ring, n),
        };
        diag_canonical.diag_canonical_update_ext(the);
        diag_canonical.invariant_factor_update_ext(the);
        return { ...the, diag_canonical };
    }
    rank() {
        let row_canonical = this.row_canonical_form();
        let rank = 0;
        for (let row of row_canonical.rows()) {
            if (row.some(v => !this.ring.zero_p(v))) {
                rank += 1;
            }
        }
        return rank;
    }
    image() {
        return this.transpose()
            .row_canonical_form()
            .transpose()
            .slice(null, [0, this.rank()]);
    }
    kernel() {
        let [m, n] = this.shape;
        let r = this.rank();
        let { col_trans } = this.diag_canonical_decomposition();
        return col_trans.slice(null, [r, n]);
    }
    solve(b) {
        let [m, n] = this.shape;
        assert_1.default(b.size === m);
        let r = this.rank();
        let { row_trans, col_trans, diag_canonical, } = this.diag_canonical_decomposition();
        let c = b.trans(row_trans);
        let vector = vector_t.ring_zeros(this.ring, n);
        for (let i of ut.range(0, c.size)) {
            let x = c.get(i);
            let y = i < r ? diag_canonical.get(i, i) : this.ring.zero;
            if (this.ring.zero_p(y)) {
                if (!this.ring.zero_p(x)) {
                    return null;
                }
            }
            else {
                let [q, r] = this.ring.divmod(x, y);
                if (this.ring.zero_p(r)) {
                    vector.set(i, q);
                }
                else {
                    return null;
                }
            }
        }
        return vector.trans(col_trans);
    }
    solve_matrix(matrix) {
        let [m, n] = this.shape;
        let [p, q] = matrix.shape;
        assert_1.default(p === m);
        let solution = matrix_t.ring_zeros(this.ring, n, q);
        for (let [i, col] of matrix.col_entries()) {
            let x = this.solve(col);
            if (x === null) {
                return null;
            }
            else {
                solution.set_col(i, x);
            }
        }
        return solution;
    }
}
exports.matrix_t = matrix_t;
class vector_t {
    constructor(the) {
        this.ring = the.ring;
        this.buffer = the.buffer;
        this.shape = the.shape;
        this.strides = the.strides;
        this.offset = the.offset === undefined ? 0 : the.offset;
        this.size = the.shape[0];
    }
    static from_ring_buffer(ring, buffer) {
        return new vector_t({
            ring,
            buffer,
            shape: [buffer.length],
            strides: [1],
        });
    }
    static from_ring_Array(ring, array) {
        return vector_t.from_ring_buffer(ring, array);
    }
    get_linear_index(x) {
        return (this.offset +
            x * this.strides[0]);
    }
    get(i) {
        return this.buffer[this.get_linear_index(i)];
    }
    set(i, v) {
        this.buffer[this.get_linear_index(i)] = v;
        return this;
    }
    toArray() {
        let array = [];
        for (let i of ut.range(0, this.size)) {
            array.push(this.get(i));
        }
        return array;
    }
    print() {
        console.log("vector:");
        console.table(this.toArray());
    }
    *indexes() {
        for (let i of ut.range(0, this.size)) {
            yield i;
        }
    }
    update_at(i, f) {
        return this.set(i, f(this.get(i)));
    }
    update(f) {
        for (let i of this.indexes()) {
            this.set(i, f(this.get(i)));
        }
        return this;
    }
    *values() {
        for (let i of ut.range(0, this.size)) {
            let v = this.buffer[this.get_linear_index(i)];
            yield v;
        }
    }
    *entries() {
        for (let i of ut.range(0, this.size)) {
            let v = this.buffer[this.get_linear_index(i)];
            yield [i, v];
        }
    }
    copy() {
        let buffer = new Array(this.size);
        let vector = vector_t.from_ring_buffer(this.ring, buffer);
        for (let [i, x] of this.entries()) {
            vector.set(i, x);
        }
        return vector;
    }
    *zip(that) {
        let this_iter = this.values();
        let that_iter = that.values();
        while (true) {
            let this_next = this_iter.next();
            let that_next = that_iter.next();
            if (this_next.done || that_next.done) {
                return;
            }
            else {
                yield [this_next.value, that_next.value];
            }
        }
    }
    eq(that) {
        if (this.size !== that.size) {
            return false;
        }
        for (let [x, y] of this.zip(that)) {
            if (x !== y) {
                return false;
            }
        }
        return true;
    }
    dot(that) {
        assert_1.default(this.size === that.size);
        let product = this.ring.zero;
        for (let [i, y] of that.entries()) {
            product = this.ring.add(product, this.ring.mul(this.get(i), y));
        }
        return product;
    }
    map(f) {
        let vector = this.copy();
        for (let [i, v] of this.entries()) {
            vector.update_at(i, f);
        }
        return vector;
    }
    scale(a) {
        return this.map(n => this.ring.mul(n, a));
    }
    update_scale(a) {
        return this.update(n => this.ring.mul(n, a));
    }
    static ring_constant(ring, n, size) {
        let buffer = new Array(size).fill(n);
        return vector_t.from_ring_buffer(ring, buffer);
    }
    static ring_zeros(ring, size) {
        return vector_t.ring_constant(ring, ring.zero, size);
    }
    static ring_ones(ring, size) {
        return vector_t.ring_constant(ring, ring.one, size);
    }
    trans(matrix) {
        let [m, n] = matrix.shape;
        assert_1.default(n === this.size);
        let vector = vector_t.ring_zeros(this.ring, m);
        for (let i of ut.range(0, m)) {
            vector.set(i, this.dot(matrix.row(i)));
        }
        return vector;
    }
    append(that) {
        let buffer = new Array(this.size + that.size);
        let vector = vector_t.from_ring_buffer(this.ring, buffer);
        for (let [i, x] of this.entries()) {
            vector.set(i, x);
        }
        for (let [i, x] of that.entries()) {
            vector.set(i + this.size, x);
        }
        return vector;
    }
    some(p) {
        for (let v of this.values()) {
            if (p(v)) {
                return true;
            }
        }
        return false;
    }
    every(p) {
        for (let v of this.values()) {
            if (!p(v)) {
                return false;
            }
        }
        return true;
    }
    update_add(that) {
        assert_1.default(this.size === that.size);
        for (let [i, x] of that.entries()) {
            this.update_at(i, v => this.ring.add(v, x));
        }
        return this;
    }
    add(that) {
        assert_1.default(this.size === that.size);
        let vector = this.copy();
        for (let [i, x] of that.entries()) {
            vector.update_at(i, v => this.ring.add(v, x));
        }
        return vector;
    }
    update_sub(that) {
        assert_1.default(this.size === that.size);
        for (let [i, x] of that.entries()) {
            this.update_at(i, v => this.ring.sub(v, x));
        }
        return this;
    }
    sub(that) {
        assert_1.default(this.size === that.size);
        let vector = this.copy();
        for (let [i, x] of that.entries()) {
            vector.update_at(i, v => this.ring.sub(v, x));
        }
        return vector;
    }
    reduce_with(init, f) {
        let acc = init;
        for (let v of this.values()) {
            acc = f(acc, v);
        }
        return acc;
    }
    reduce(f) {
        assert_1.default(this.size > 0);
        if (this.size === 1) {
            return this.get(0);
        }
        else {
            let acc = this.get(0);
            for (let i of ut.range(1, this.size)) {
                acc = f(acc, this.get(i));
            }
            return acc;
        }
    }
    argfirst(p) {
        let lo = 0;
        let hi = this.size;
        return argfirst(lo, hi, i => p(this.get(i)));
    }
    first(p) {
        let arg = this.argfirst(p);
        if (arg === null) {
            return null;
        }
        else {
            return this.get(arg);
        }
    }
    invariant_factors_p() {
        for (let i of ut.range(0, this.size)) {
            let x = this.get(i);
            if (!this.ring.zero_p(x)) {
                for (let k of ut.range(i + 1, this.size)) {
                    let y = this.get(k);
                    let [q, r] = this.ring.divmod(y, x);
                    if (!this.ring.zero_p(r)) {
                        return false;
                    }
                }
            }
        }
        return true;
    }
    zero_p() {
        return this.every(x => this.ring.zero_p(x));
    }
    one_p() {
        return this.every(x => this.ring.one_p(x));
    }
}
exports.vector_t = vector_t;
//# sourceMappingURL=euclid.js.map