"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
class vertex_t {
    constructor(id, info) {
        this.id = id;
        this.info = info;
    }
    // - shallow copy
    copy() {
        return new vertex_t(this.id, this.info);
    }
    more_info(info) {
        return new vertex_t(this.id, {
            ...this.info,
            ...info,
        });
    }
}
exports.vertex_t = vertex_t;
class edge_t {
    constructor(id, start, end, info) {
        this.id = id;
        this.start = start;
        this.end = end;
        this.info = info;
    }
    // - shallow copy
    copy() {
        return new edge_t(this.id, this.start, this.end, this.info);
    }
    more_info(info) {
        return new edge_t(this.id, this.start, this.end, {
            ...this.info,
            ...info,
        });
    }
}
exports.edge_t = edge_t;
class graph_t {
    // simple graph representation
    // - in the following "O" notation V and E denote sizes
    // - space ~ O (V + E)
    constructor(vertex_map = new Map(), edge_map = new Map()) {
        this.vertex_map = vertex_map;
        this.edge_map = edge_map;
    }
    // self-builder
    // - side-effect on existing object under the same id
    vertex(v, info) {
        if (this.vertex_map.has(v))
            throw new Error(`[error] graph_t.vertex fail`);
        else
            this.vertex_map.set(v, new vertex_t(v, info));
        return this;
    }
    edge(e, start, end, info) {
        if (this.edge_map.has(e))
            throw new Error(`[error] graph_t.edge fail`);
        else
            this.edge_map.set(e, new edge_t(e, start, end, info));
        return this;
    }
    // - shallow copy
    // - time ~ O (V + E)
    copy() {
        let graph = new graph_t();
        this.vertex_map.forEach((vertex, id) => {
            graph.vertex_map.set(id, vertex);
        });
        this.edge_map.forEach((edge, id) => {
            graph.edge_map.set(id, edge);
        });
        return graph;
    }
    // - time ~ O (E)
    // - with side-effect on the set
    connected_vertex_set(vertex_set) {
        for (let edge of this.edge_map.values()) {
            if (vertex_set.has(edge.start))
                vertex_set.add(edge.end);
            else if (vertex_set.has(edge.end))
                vertex_set.add(edge.start);
        }
        return vertex_set;
    }
    // - time ~ O (E)
    connected_vertex_set_of_vertex(v) {
        return this.connected_vertex_set(new Set([v]));
    }
    // - time ~ O (E)
    connected_vertex_p(v1, v2) {
        return this
            .connected_vertex_set_of_vertex(v1)
            .has(v2);
    }
    // - time ~ O (V + E)
    subgraph_of_vertex_set(vertex_set) {
        let subgraph = new graph_t();
        this.vertex_map.forEach((vertex, id) => {
            if (vertex_set.has(id))
                subgraph.vertex_map.set(id, vertex);
        });
        this.edge_map.forEach((edge, id) => {
            if (vertex_set.has(edge.start) ||
                vertex_set.has(edge.end))
                subgraph.edge_map.set(id, edge);
        });
        return subgraph;
    }
    // - connected component is max connected subgraph
    // - time ~ O (V + E)
    connected_component_of_vertex(v) {
        return this.subgraph_of_vertex_set(this.connected_vertex_set_of_vertex(v));
    }
}
exports.graph_t = graph_t;
//# sourceMappingURL=graph-info.js.map