"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Normalization by Evaluation:
 *   Dependent Types and Impredicativity
 * by Andreas Abel
 */
/**
 * Chapter 2:
 *   Simple Types: From Evaluation to Normalization
 */
class type_t {
}
class nat_t extends type_t {
    eq(that) {
        return that instanceof nat_t;
    }
}
class arrow_t extends type_t {
    constructor(arg_type, ret_type) {
        super();
        this.arg_type = arg_type;
        this.ret_type = ret_type;
    }
    eq(that) {
        return that instanceof arrow_t
            && this.arg_type.eq(that.arg_type)
            && this.ret_type.eq(that.ret_type);
    }
}
class ctx_t {
    constructor(map) {
        this.map = map;
    }
    static empty_context() {
        return new ctx_t(new Map());
    }
    ext(name, t) {
        if (this.map.has(name)) {
            throw new Error(`name: ${name} already exists in this context`);
        }
        return new ctx_t(new Map(this.map).set(name, t));
    }
}
/**
 * Constant value of type `this.T`.
 */
class const_t {
}
class zero_t extends const_t {
    constructor() {
        super();
        this.T = new nat_t();
    }
    eq(that) {
        return that instanceof zero_t;
    }
}
class suc_t extends const_t {
    constructor() {
        super();
        this.T = new arrow_t(new nat_t(), new nat_t());
    }
    eq(that) {
        return that instanceof suc_t;
    }
}
/**
 * Primitive recursion into type `this.into_type`.
 */
class rec_t extends const_t {
    constructor(into_type) {
        super();
        this.into_type = into_type;
        this.T = new arrow_t(this.into_type, new arrow_t(new arrow_t(new nat_t(), new arrow_t(this.into_type, this.into_type)), new arrow_t(new nat_t(), this.into_type)));
    }
    eq(that) {
        return that instanceof rec_t
            && this.into_type.eq(that.into_type);
    }
}
/**
 * Well-typed terms, in context.
 * - We need a class `term_t`
 *   and a `well_typed` judgment method on `term_t`.
 * - `term_t` does not have a `T` field,
 *   for the type of a term depends context.
 */
class term_t {
}
class constant_t extends term_t {
    constructor(c) {
        super();
        this.c = c;
    }
    eq(that) {
        return that instanceof constant_t
            && this.c.eq(that.c);
    }
    well_typed_p(_ctx, T) {
        return this.c.T.eq(T);
    }
    normal_p(ctx, T) {
        if (this.neutral_p(ctx, T)) {
            return true;
        }
        else if (this.c instanceof zero_t) {
            return this.well_typed_p(ctx, T);
        }
        else {
            return false;
        }
    }
    neutral_p(ctx, T) {
        return false;
    }
}
class variable_t extends term_t {
    constructor(name) {
        super();
        this.name = name;
    }
    eq(that) {
        return that instanceof variable_t
            && this.name === that.name;
    }
    well_typed_p(ctx, T) {
        let X = ctx.map.get(name);
        if (X !== undefined && (X.eq(T))) {
            return true;
        }
        else {
            return false;
        }
    }
    normal_p(ctx, T) {
        if (this.neutral_p(ctx, T)) {
            return true;
        }
        else {
            return false;
        }
    }
    neutral_p(ctx, T) {
        return this.well_typed_p(ctx, T);
    }
}
class lambda_t extends term_t {
    constructor(name, term) {
        super();
        this.name = name;
        this.term = term;
    }
    eq(that) {
        return that instanceof lambda_t
            && this.name === that.name
            && this.term === that.term;
    }
    well_typed_p(ctx, T) {
        if (T instanceof arrow_t) {
            return this.term.well_typed_p(ctx.ext(this.name, T.arg_type), T.ret_type);
        }
        else {
            return false;
        }
    }
    normal_p(ctx, T) {
        if (this.neutral_p(ctx, T)) {
            return true;
        }
        else if (T instanceof arrow_t) {
            return this.term.normal_p(ctx.ext(this.name, T.arg_type), T.ret_type);
        }
        else {
            return false;
        }
    }
    neutral_p(ctx, T) {
        return false;
    }
}
class apply_t extends term_t {
    constructor(fun, arg, arg_type) {
        super();
        this.fun = fun;
        this.arg = arg;
        this.arg_type = arg_type;
    }
    eq(that) {
        return that instanceof apply_t
            && this.fun === that.fun
            && this.arg === that.arg
            && this.arg_type === that.arg_type;
    }
    well_typed_p(ctx, T) {
        return this.arg.well_typed_p(ctx, this.arg_type)
            && this.fun.well_typed_p(ctx, new arrow_t(this.arg_type, T));
    }
    normal_p(ctx, T) {
        if (this.neutral_p(ctx, T)) {
            return true;
        }
        else if (T.eq(new nat_t()) &&
            this.fun instanceof constant_t &&
            this.fun.c instanceof suc_t) {
            return this.arg.normal_p(ctx, new nat_t());
        }
        else {
            return false;
        }
    }
    neutral_p(ctx, T) {
        if (this.arg.normal_p(ctx, this.arg_type) &&
            this.fun.neutral_p(ctx, new arrow_t(this.arg_type, T))) {
            return true;
        }
        else if (this.arg.neutral_p(ctx, new nat_t()) &&
            this.fun instanceof apply_t &&
            this.fun.arg.normal_p(ctx, new arrow_t(new nat_t(), new arrow_t(T, T))) &&
            this.fun.fun instanceof apply_t &&
            this.fun.fun.arg.normal_p(ctx, T) &&
            this.fun.fun.fun instanceof constant_t &&
            this.fun.fun.fun.c instanceof rec_t) {
            return true;
        }
        else {
            return false;
        }
    }
}
/**
 * Definitional Equality (a.k.a. sameness in "the little typer")
 * - not complete -- does not capture the semantic equality
 * - but decidable -- we can write a checker for this equality
 */
class definitional_equality_t {
    constructor(x, y) {
        this.x = x;
        this.y = y;
    }
    swap() {
        return new definitional_equality_t(this.y, this.x);
    }
    check_same(ctx, T) {
        return this.check_same_directed(ctx, T)
            || this.swap().check_same_directed(ctx, T);
    }
    check_same_directed(ctx, T) {
        return this.reflexivity(ctx, T)
            || this.beta_equality(ctx, T)
            || this.eta_equality(ctx, T)
            || this.compatibility(ctx, T);
    }
    reflexivity(ctx, T) {
        return this.x.eq(this.y)
            && this.x.well_typed_p(ctx, T);
    }
    // TODO handle lambda applied to arg
    beta_equality(ctx, T) {
        if (this.x instanceof apply_t &&
            this.x.arg.eq(new constant_t(new zero_t())) &&
            this.x.fun instanceof apply_t &&
            this.x.fun.fun instanceof apply_t &&
            this.x.fun.fun.fun instanceof constant_t &&
            this.x.fun.fun.fun.c instanceof rec_t &&
            this.y.eq(this.x.fun.fun.arg) &&
            this.x.fun.arg.well_typed_p(ctx, new arrow_t(new nat_t(), new arrow_t(T, T))) &&
            this.x.fun.fun.arg.well_typed_p(ctx, T)) {
            return true;
        }
        else if (this.x instanceof apply_t &&
            this.x.arg instanceof apply_t &&
            this.x.arg.fun.eq(new constant_t(new suc_t())) &&
            this.x.arg.arg.well_typed_p(ctx, new nat_t()) &&
            this.x.fun instanceof apply_t &&
            this.x.fun.fun instanceof apply_t &&
            this.x.fun.fun.fun instanceof constant_t &&
            this.x.fun.fun.fun.c instanceof rec_t &&
            this.y.eq(this.x.fun.fun.arg) &&
            this.x.fun.arg.well_typed_p(ctx, new arrow_t(new nat_t(), new arrow_t(T, T))) &&
            this.x.fun.fun.arg.well_typed_p(ctx, T) &&
            this.y instanceof apply_t &&
            this.y.fun instanceof apply_t &&
            this.y.fun.fun.eq(this.x.fun.arg) &&
            this.y.fun.arg.eq(this.x.fun.fun.arg) &&
            this.y.arg instanceof apply_t &&
            this.y.arg.fun instanceof apply_t &&
            this.y.arg.fun.fun instanceof apply_t &&
            this.y.arg.arg.eq(this.x.arg.arg) &&
            this.y.arg.fun.arg.eq(this.x.fun.arg) &&
            this.y.arg.fun.fun.arg.eq(this.x.fun.fun.arg) &&
            this.y.arg.fun.fun.fun instanceof constant_t &&
            this.y.arg.fun.fun.fun.c instanceof rec_t) {
            return true;
        }
        else {
            return false;
        }
    }
    eta_equality(ctx, T) {
        if (this.x instanceof lambda_t &&
            this.x.term instanceof apply_t &&
            this.x.term.arg instanceof variable_t &&
            this.x.term.arg.name === this.x.name &&
            this.x.term.fun.eq(this.y) &&
            T instanceof arrow_t &&
            this.y.well_typed_p(ctx, T)) {
            return true;
        }
        else {
            return false;
        }
    }
    compatibility(ctx, T) {
        if (this.x.eq(this.y) &&
            this.x.well_typed_p(ctx, T)) {
            return true;
        }
        else if (this.x instanceof lambda_t &&
            this.y instanceof lambda_t &&
            T instanceof arrow_t &&
            this.x.name === this.y.name) {
            return new definitional_equality_t(this.x.term, this.y.term).check_same(ctx.ext(this.x.name, T.arg_type), T.ret_type);
        }
        else if (this.x instanceof apply_t &&
            this.y instanceof apply_t &&
            this.x.arg_type.eq(this.y.arg_type)) {
            return new definitional_equality_t(this.x.fun, this.y.fun).check_same(ctx, new arrow_t(this.x.arg_type, T)) && new definitional_equality_t(this.x.arg, this.y.arg).check_same(ctx, this.x.arg_type);
        }
        else {
            return false;
        }
    }
}
function reflection(ctx, T, u) {
    if (!u.neutral_p(ctx, T)) {
        throw new Error("! u.neutral_p (ctx, T)");
    }
    if (T instanceof nat_t) {
        return u;
    }
    else if (T instanceof arrow_t) {
        let v = "TODO";
        return reflection(ctx, T.ret_type, new apply_t(u, v, T.ret_type));
    }
    else {
        throw new Error("TODO");
    }
}
function reification(ctx, T, u) {
    throw new Error("TODO");
}
//# sourceMappingURL=simple.js.map