"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
const _ = __importStar(require("lodash"));
const assert_1 = __importDefault(require("assert"));
const ut = __importStar(require("./util"));
/**
 * strides based row-major ndarray of number
 */
class array_t {
    /**
     * I keep the basic constructor low level,
     * and mainly use specific constructors in client modules.
     */
    constructor(buffer, shape, strides, offset = 0) {
        this.buffer = buffer;
        this.shape = shape;
        this.strides = strides;
        this.offset = offset;
        this.order = shape.length;
        this.size = array_t.shape_to_size(shape);
        if (strides.length !== shape.length) {
            throw new Error("strides shape length mismatch");
        }
        if (buffer.length < this.size + offset) {
            throw new Error("buffer not large enough");
        }
    }
    static shape_to_size(shape) {
        return shape.reduce((acc, cur) => acc * cur);
    }
    static shape_to_strides(shape) {
        let strides = [];
        let acc = 1;
        shape.slice().reverse().forEach((x) => {
            strides.push(acc);
            acc *= x;
        });
        return strides.reverse();
    }
    get_linear_index(index) {
        if (index.length !== this.shape.length) {
            throw new Error("index length mismatch");
        }
        let linear_index = this.offset;
        for (let i = 0; i < index.length; i += 1) {
            linear_index += index[i] * this.strides[i];
        }
        return linear_index;
    }
    linear_get(i) {
        return this.buffer[i];
    }
    linear_set(i, v) {
        this.buffer[i] = v;
        return this;
    }
    get(index) {
        let i = this.get_linear_index(index);
        return this.linear_get(i);
    }
    set(index, x) {
        let i = this.get_linear_index(index);
        this.linear_set(i, x);
        return this;
    }
    update_at(index, f) {
        return this.set(index, f(this.get(index)));
    }
    *values() {
        for (let index of indexes_of_shape(this.shape)) {
            yield this.get(index);
        }
    }
    *entries() {
        for (let index of indexes_of_shape(this.shape)) {
            yield [
                index.slice(),
                this.get(index),
            ];
        }
    }
    *indexes() {
        for (let index of indexes_of_shape(this.shape)) {
            yield index.slice();
        }
    }
    reduce_with(init, f) {
        let acc = init;
        for (let v of this.values()) {
            acc = f(acc, v);
        }
        return acc;
    }
    copy() {
        let buffer = new Float64Array(this.size);
        let array = new array_t(buffer, this.shape, array_t.shape_to_strides(this.shape));
        for (let [i, x] of this.entries()) {
            array.set(i, x);
        }
        return array;
    }
    proj(index) {
        if (index.length !== this.shape.length) {
            throw new Error("index length mismatch");
        }
        let shape = new Array();
        let strides = new Array();
        let offset = this.offset;
        for (let [k, v] of index.entries()) {
            if (v === null) {
                shape.push(this.shape[k]);
                strides.push(this.strides[k]);
            }
            else {
                offset += v * this.strides[k];
            }
        }
        return new array_t(this.buffer, shape, strides, offset);
    }
    slice(index) {
        if (index.length !== this.shape.length) {
            throw new Error("index length mismatch");
        }
        let shape = this.shape.slice();
        let offset = this.offset;
        for (let [k, v] of index.entries()) {
            if (v !== null) {
                let [start, end] = v;
                shape[k] = end - start;
                offset += start * this.strides[k];
            }
        }
        return new array_t(this.buffer, shape, this.strides, offset);
    }
    /**
     * the order of `src` array can be higher or lower,
     * as long as its `.values` match `tar.indexes`.
     */
    put(index, src) {
        let tar = this.slice(index);
        let index_array = Array.from(tar.indexes());
        let value_array = Array.from(src.values());
        if (index_array.length !== value_array.length) {
            throw new Error("size mismatch");
        }
        for (let k in index_array) {
            let i = index_array[k];
            let v = value_array[k];
            tar.set(i, v);
        }
        return this;
    }
    put_porj(index, src) {
        let slice_index = index.map(i => {
            if (i === null) {
                return null;
            }
            else {
                return [i, i + 1];
            }
        });
        return this.put(slice_index, src);
    }
    add(that) {
        let array = this.copy();
        for (let [i, v] of that.entries()) {
            array.update_at(i, x => x + v);
        }
        return array;
    }
    static from_1darray(array) {
        let buffer = Float64Array.from(array);
        let shape = [array.length];
        let strides = array_t.shape_to_strides(shape);
        return new array_t(buffer, shape, strides);
    }
    to_1darray() {
        return Array.from(this.values());
    }
    static from_2darray(array) {
        let y_length = array.length;
        let x_length = array[0].length;
        for (let a of array) {
            if (a.length !== x_length) {
                throw new Error("inner array length mismatch");
            }
        }
        let buffer = Float64Array.from(array.flat());
        let shape = [y_length, x_length];
        let strides = array_t.shape_to_strides(shape);
        return new array_t(buffer, shape, strides);
    }
    to_2darray() {
        let array = [];
        let [x, _] = this.shape;
        for (let i = 0; i < x; i++) {
            array.push(this.proj([i, null]).to_1darray());
        }
        return array;
    }
    static from_3darray(array) {
        let z_length = array.length;
        let y_length = array[0].length;
        let x_length = array[0][0].length;
        for (let a of array) {
            if (a.length !== y_length) {
                throw new Error("inner array length mismatch");
            }
            else {
                for (let b of a) {
                    if (b.length !== x_length) {
                        throw new Error("inner inner array length mismatch");
                    }
                }
            }
        }
        let buffer = Float64Array.from(array.flat(2));
        let shape = [z_length, y_length, x_length];
        let strides = array_t.shape_to_strides(shape);
        return new array_t(buffer, shape, strides);
    }
    static from_buffer(buffer, shape) {
        let strides = array_t.shape_to_strides(shape);
        return new array_t(buffer, shape, strides);
    }
    static constant(n, shape) {
        let size = array_t.shape_to_size(shape);
        let buffer = new Float64Array(size);
        buffer.fill(n);
        let strides = array_t.shape_to_strides(shape);
        return new array_t(buffer, shape, strides);
    }
    static zeros(shape) {
        return array_t.constant(0, shape);
    }
    static ones(shape) {
        return array_t.constant(1, shape);
    }
    fill(x) {
        this.buffer.fill(x);
        return this;
    }
    static proj_index_max_p(index, shape) {
        if (index.length !== shape.length) {
            throw new Error("index length mismatch");
        }
        let flag = true;
        for (let k in index) {
            let i = index[k];
            let s = shape[k];
            if (i !== null && i < (s - 1)) {
                flag = false;
            }
        }
        return flag;
    }
    static proj_index_inc_with_shape(index, shape) {
        if (index.length !== shape.length) {
            throw new Error("index length mismatch");
        }
        if (array_t.proj_index_max_p(index, shape)) {
            throw new Error("index out of shape");
        }
        let [i] = index.slice(-1);
        let [s] = shape.slice(-1);
        if (i === null) {
            let new_index = index.slice();
            let new_shape = shape.slice();
            new_index.pop();
            new_shape.pop();
            new_index = array_t.proj_index_inc_with_shape(new_index, new_shape);
            new_index.push(null);
            return new_index;
        }
        else if (i >= s - 1) {
            let new_index = index.slice();
            let new_shape = shape.slice();
            new_index.pop();
            new_shape.pop();
            new_index = array_t.proj_index_inc_with_shape(new_index, new_shape);
            new_index.push(0);
            return new_index;
        }
        else {
            let new_index = index.slice();
            new_index.pop();
            new_index.push(i + 1);
            return new_index;
        }
    }
    print() {
        if (this.order === 1) {
            console.table(this.to_1darray());
        }
        else if (this.order === 2) {
            console.table(this.to_2darray());
        }
        else {
            let index = this.shape.slice().fill(0);
            index.pop();
            index.pop();
            index.push(null);
            index.push(null);
            while (true) {
                console.log("array index:", index);
                this.proj(index).print();
                index =
                    array_t.proj_index_inc_with_shape(index, this.shape);
                if (array_t.proj_index_max_p(index, this.shape)) {
                    console.log("array index:", index);
                    this.proj(index).print();
                    return;
                }
            }
        }
    }
    *zip(that) {
        let this_iter = this.values();
        let that_iter = that.values();
        while (true) {
            let this_next = this_iter.next();
            let that_next = that_iter.next();
            if (this_next.done || that_next.done) {
                return;
            }
            else {
                yield [this_next.value, that_next.value];
            }
        }
    }
    eq(that) {
        if (this.size !== that.size) {
            return false;
        }
        if (this.order !== that.order) {
            return false;
        }
        if (!_.isEqual(this.shape, that.shape)) {
            return false;
        }
        for (let [x, y] of this.zip(that)) {
            if (x !== y) {
                return false;
            }
        }
        return true;
    }
    map(f) {
        let buffer = new Float64Array(this.size);
        let array = new array_t(buffer, this.shape, array_t.shape_to_strides(this.shape));
        for (let [i, x] of this.entries()) {
            array.set(i, f(x));
        }
        return array;
    }
    for_each(f) {
        for (let x of this.values()) {
            f(x);
        }
        return this;
    }
    append(k, that) {
        assert_1.default(this.order === that.order);
        for (let j of ut.range(0, this.order)) {
            if (j !== k) {
                assert_1.default(this.shape[j] === that.shape[j]);
            }
        }
        let shape = this.shape.slice();
        shape[k] = this.shape[k] + that.shape[k];
        let buffer = new Float64Array(array_t.shape_to_size(shape));
        let strides = array_t.shape_to_strides(shape);
        let array = new array_t(buffer, shape, strides);
        let offset = this.shape[k];
        for (let i of array.indexes()) {
            if (i[k] < offset) {
                array.set(i, this.get(i));
            }
            else {
                let j = i.slice();
                j[k] = j[k] - offset;
                array.set(i, that.get(j));
            }
        }
        return array;
    }
    reshape(permutation) {
        let shape = new Array();
        let strides = new Array();
        for (let i of permutation) {
            shape.push(this.shape[i]);
            strides.push(this.strides[i]);
        }
        return new array_t(this.buffer, shape, strides, this.offset);
    }
    permute(k, permutation) {
        assert_1.default(permutation.size === this.shape[k]);
        let array = this.copy();
        for (let i of this.indexes()) {
            let j = i.slice();
            j[k] = permutation.get(j[k]);
            array.set(j, this.get(i));
        }
        return array;
    }
    // [1, 2, <3>, 4]
    //    [2, <3>, 4, 5]
    // => [1, 2, 4, 2, 4, 5]
    contract(that, i, j) {
        let this_size = this.shape[i];
        let that_size = that.shape[j];
        if (this_size !== that_size) {
            throw new Error("size mismatch");
        }
        let shape = new Array();
        this.shape.forEach((s, k) => {
            if (k !== i) {
                shape.push(s);
            }
        });
        that.shape.forEach((s, k) => {
            if (k !== j) {
                shape.push(s);
            }
        });
        let size = array_t.shape_to_size(shape);
        let buffer = new Float64Array(size);
        let strides = array_t.shape_to_strides(shape);
        let array = new array_t(buffer, shape, strides);
        let split_index = (index) => {
            let left = index.slice(0, this.order - 1);
            left.splice(i, 0, null);
            let right = index.slice(this.order - 1);
            right.splice(j, 0, null);
            return [left, right];
        };
        for (let index of array.indexes()) {
            let [left, right] = split_index(index);
            let sum = 0;
            let zip = this.proj(left).zip(that.proj(right));
            for (let [x, y] of zip) {
                sum += x * y;
            }
            array.set(index, sum);
        }
        return array;
    }
    static from_lower_order(lower) {
        assert_1.default(lower.length !== 0);
        let first_array = lower[0];
        let first_shape = first_array.shape;
        let shape = [lower.length].concat(first_array.shape);
        let size = array_t.shape_to_size(shape);
        let buffer = new Float64Array(size);
        let higher = new array_t(buffer, shape, array_t.shape_to_strides(shape));
        for (let i of ut.range(0, lower.length)) {
            let array = lower[i];
            assert_1.default(_.isEqual(array.shape, first_shape));
            let index = [i];
            index = index.concat(ut.repeats(null, shape.length - 1));
            higher.put_porj(index, array);
        }
        return higher;
    }
}
exports.array_t = array_t;
function array1d_p(array) {
    if (array.length === 0) {
        return true;
    }
    else {
        return typeof array[0] === 'number';
    }
}
exports.array1d_p = array1d_p;
function array2d_p(array) {
    if (array.length === 0) {
        return true;
    }
    else {
        return array1d_p(array[0]);
    }
}
exports.array2d_p = array2d_p;
function array3d_p(array) {
    if (array.length === 0) {
        return true;
    }
    else {
        return array2d_p(array[0]);
    }
}
exports.array3d_p = array3d_p;
function array(array) {
    if (array1d_p(array)) {
        return array_t.from_1darray(array);
    }
    else if (array2d_p(array)) {
        return array_t.from_2darray(array);
    }
    else if (array3d_p(array)) {
        return array_t.from_3darray(array);
    }
    else {
        throw new Error("can only handle Array1d | Array2d | Array3d");
    }
}
exports.array = array;
function index_max_p(index, shape) {
    for (let k in index) {
        let i = index[k];
        let s = shape[k];
        if (i < s - 1) {
            return false;
        }
    }
    return true;
}
exports.index_max_p = index_max_p;
/**
 * recursive side-effect over index
 */
function index_step(index, shape, cursor) {
    let i = index[cursor];
    let s = shape[cursor];
    if (i < s - 1) {
        index[cursor] = i + 1;
        return index;
    }
    else {
        index[cursor] = 0;
        return index_step(index, shape, cursor - 1);
    }
}
exports.index_step = index_step;
function* indexes_of_shape(shape) {
    let size = shape.length;
    let index = new Array(size).fill(0);
    yield index;
    while (true) {
        if (index_max_p(index, shape)) {
            return;
        }
        else {
            yield index_step(index, shape, size - 1);
        }
    }
}
exports.indexes_of_shape = indexes_of_shape;
//# sourceMappingURL=ndarray.js.map