"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const assert_1 = __importDefault(require("assert"));
const eu = __importStar(require("./euclid"));
const set_1 = require("./abstract/set");
exports.config = {
    epsilon: 0.000000001
};
function epsilon_p(x) {
    return Math.abs(x) < exports.config.epsilon;
}
exports.epsilon_p = epsilon_p;
function non_epsilon_p(x) {
    return Math.abs(x) >= exports.config.epsilon;
}
exports.non_epsilon_p = non_epsilon_p;
exports.nums = new set_1.set_t({
    eq: (x, y) => x === y
});
exports.ring = eu.ring({
    elements: exports.nums,
    zero: 0,
    add: (x, y) => x + y,
    neg: (x) => -x,
    one: 1,
    mul: (x, y) => x * y,
    degree_lt: (x, y) => false,
    divmod: (x, y) => [x / y, 0],
});
function argmax(lo, hi, f) {
    let arg = lo;
    let cur = f(lo);
    for (let i = lo; i < hi; i++) {
        let next = f(i);
        if (cur < next) {
            arg = i;
            cur = next;
        }
    }
    return arg;
}
exports.argmax = argmax;
class matrix_t extends eu.matrix_t {
    constructor(the) {
        super({
            ...the,
            ring: exports.ring,
        });
    }
    copy() {
        return new matrix_t(super.copy());
    }
    static constant(n, x, y) {
        return new matrix_t(eu.matrix_t.ring_constant(exports.ring, n, x, y));
    }
    static zeros(x, y) {
        return new matrix_t(eu.matrix_t.ring_zeros(exports.ring, x, y));
    }
    static ones(x, y) {
        return new matrix_t(eu.matrix_t.ring_ones(exports.ring, x, y));
    }
    static id(n) {
        return new matrix_t(eu.matrix_t.ring_id(exports.ring, n));
    }
    transpose() {
        return new matrix_t(super.transpose());
    }
    append_cols(that) {
        return new matrix_t(this.append_cols(that));
    }
    row_echelon_form() {
        let matrix = this.copy();
        let [m, n] = this.shape;
        let h = 0; // init pivot row
        let k = 0; // init pivot column
        while (h < m && k < n) {
            // find the next pivot
            let piv = argmax(h, m, (i) => Math.abs(matrix.get(i, k)));
            if (epsilon_p(matrix.get(piv, k))) {
                // no pivot in this column, pass to next column
                k += 1;
            }
            else {
                if (h !== piv) {
                    matrix.update_swap_rows(h, piv);
                }
                // for all rows below pivot
                for (let i = h + 1; i < m; i++) {
                    let f = matrix.get(i, k) / matrix.get(h, k);
                    matrix.set(i, k, 0);
                    // for all remaining elements in current row
                    for (let j = k + 1; j < n; j++) {
                        let v = matrix.get(i, j) - matrix.get(h, j) * f;
                        matrix.set(i, j, v);
                    }
                }
                h += 1;
                k += 1;
            }
        }
        return matrix;
    }
    /**
     * with all pivots equal to 1.
     */
    unit_row_echelon_form() {
        let matrix = this.row_echelon_form();
        matrix.for_each_row_index((row, i) => {
            let pivot = row.first(x => !epsilon_p(x));
            if (pivot !== null) {
                row.update_scale(1 / pivot);
            }
        });
        return matrix;
    }
    /**
     * unit row echelon form + back substitution
  
     * The reduced row echelon form of a matrix is unique
     * i.e. does not depend on the algorithm used to compute it.
     */
    reduced_row_echelon_form() {
        let matrix = this.unit_row_echelon_form();
        for (let [i, row] of matrix.row_entries()) {
            for (let [j, sub] of matrix.row_entries()) {
                if (j > i) {
                    let arg = sub.argfirst(x => x === 1);
                    if (arg !== null) {
                        let x = matrix.get(i, arg);
                        if (!epsilon_p(x)) {
                            row.update_add(sub.scale(-x));
                        }
                    }
                }
            }
        }
        return matrix;
    }
    /**
     * P * A = L * U,
     * `permu.mul (this) .eq (lower.mul (upper))`,
     * (singular matrixes allowed)
     */
    lower_upper_decomposition() {
        let matrix = this.copy();
        let [m, n] = this.shape;
        assert_1.default(m === n);
        let record = matrix_t.zeros(m, n);
        let permu = matrix_t.id(n);
        let h = 0; // init pivot row
        let k = 0; // init pivot column
        let inver = 0;
        while (h < m && k < n) {
            // find the next pivot
            let piv = argmax(h, m, (i) => Math.abs(matrix.get(i, k)));
            if (epsilon_p(matrix.get(piv, k))) {
                // no pivot in this column, pass to next column
                k += 1;
            }
            else {
                if (h !== piv) {
                    matrix.update_swap_rows(h, piv);
                    record.update_swap_rows(h, piv);
                    permu.update_swap_rows(h, piv);
                    inver += 1;
                }
                // for all rows below pivot
                for (let i = h + 1; i < m; i++) {
                    let f = matrix.get(i, k) / matrix.get(h, k);
                    matrix.set(i, k, 0);
                    record.update_at(i, k, v => v + f);
                    // for all remaining elements in current row
                    for (let j = k + 1; j < n; j++) {
                        let v = matrix.get(i, j) - matrix.get(h, j) * f;
                        matrix.set(i, j, v);
                    }
                }
                h += 1;
                k += 1;
            }
        }
        return {
            lower: new matrix_t(record.update_add(matrix_t.id(n))),
            upper: matrix,
            permu,
            inver,
        };
    }
}
exports.matrix_t = matrix_t;
function matrix(array) {
    let new_array = new Array();
    for (let row of array) {
        let new_row = new Array();
        for (let x of row) {
            new_row.push(Number(x));
        }
        new_array.push(new_row);
    }
    return new matrix_t(eu.matrix_t.from_ring_Array2d(exports.ring, new_array));
}
exports.matrix = matrix;
class vector_t extends eu.vector_t {
    constructor(the) {
        super({
            ...the,
            ring: exports.ring,
        });
    }
    copy() {
        return new vector_t(super.copy());
    }
    static constant(n, size) {
        return new vector_t(eu.vector_t.ring_constant(exports.ring, n, size));
    }
    static zeros(size) {
        return new vector_t(eu.vector_t.ring_zeros(exports.ring, size));
    }
    static ones(size) {
        return new vector_t(eu.vector_t.ring_ones(exports.ring, size));
    }
}
exports.vector_t = vector_t;
function vector(array) {
    return new vector_t(eu.vector_t.from_ring_Array(exports.ring, array.map(Number)));
}
exports.vector = vector;
//# sourceMappingURL=num.js.map